--
-- Copyright (c) 2021-2022 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
--
-- Permission is hereby granted, free of charge, to any person obtaining a copy
-- of this software and associated documentation files (the "Software"), to deal
-- in the Software without restriction, including without limitation the rights
-- to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the Software is
-- furnished to do so, subject to the following conditions:
--
-- The above copyright notice and this permission notice shall be included in
-- all copies or substantial portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-- IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-- FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-- AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-- LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-- OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
-- SOFTWARE.
--

-- This contains <hash key, dummy value> so we can quickly check for
-- duplicates.
local dup_table = {}

local function hash(qso, txref, rxref)
	local rxstate = qso.rx.state
	local band = (qso.tx.band ~= nil) and (qso.tx.band / 1000) or "NONE"
	local mode = qso.tx.mode

	return string.format("%04u%02u%02u_%s-%s-%s_%s-%s-%s_%s-%s",
			     -- when
			     qso.start_time.year,
			     qso.start_time.month,
			     qso.start_time.day,
			     -- me
			     qso.tx.station_call,
			     qso.tx.state,
			     txref,
			     -- other op
			     qso.rx.station_call,
			     (rxstate ~= nil and rxref ~= nil) and rxstate or "NONE",
			     (rxref ~= nil) and rxref or "NONE",
			     -- contact
			     band,
			     mode)
end

local function get_hashes(qso)
	local txrefs = hlog.utils.split_on_commas(qso.additional['tx_pota'])
	local rxrefs = hlog.utils.split_on_commas(qso.additional['rx_pota'])

	local out = {}
	for _, txref in ipairs(txrefs) do
		if rxrefs == nil then
			-- regular hunter
			table.insert(out, hash(qso, txref, nil))
		else
			-- Park to Park contact, handle the other activator
			-- being in multiple entities concurrently
			for _, rxref in ipairs(rxrefs) do
				table.insert(out, hash(qso, txref, rxref))
			end
		end
	end

	return out
end

local function qso_done(qso)
	for _, h in ipairs(get_hashes(qso)) do
		dup_table[h] = true
	end
end

local function isdup(qso)
	local hashes = get_hashes(qso)
	for _, h in ipairs(hashes) do
		if dup_table[h] == nil then
			-- at least one hash doesn't exist in the list of
			-- known QSOs
			return false
		end
	end

	return true
end

local function copy_if_not_nil(dst, dkey, src)
	if src ~= nil then
		dst[dkey] = src
	end
end

-- copy country data into QSO
--
-- If the station callsign is nil (i.e., empty), we clear the various
-- QSO location fields.  In theory, we could clear the fields whenever we
-- get a lookup failure, but that seems a bit dangerous.
local function sync_zones(qso)
	local call = qso.rx.station_call
	local state = qso.rx.state

	if call ~= nil then
		local cty = hlog.country.lookup(call, state)
		if cty == nil then
			return
		end

		copy_if_not_nil(qso.rx, 'dxcc', cty.dxcc)
		copy_if_not_nil(qso.rx, 'itu', cty.itu)
		copy_if_not_nil(qso.rx, 'cqz', cty.cqz)
		copy_if_not_nil(qso.rx, 'country', cty.country)
		copy_if_not_nil(qso.rx, 'continent', cty.continent)
	else
		qso.rx.dxcc = nil
		qso.rx.itu = nil
		qso.rx.cqz = nil
		qso.rx.country = nil
		qso.rx.continent = nil
	end
end

local function field_changed(qso, fields, field)
	local value = fields[field]

	if field == "rx.station_call" then
		qso.rx.station_call = value
		sync_zones(qso)
	elseif field == "tx.rst" then
		qso.tx.rst = value
	elseif field == "rx.rst" then
		qso.rx.rst = value
	elseif field == "rx.state" then
		qso.rx.state = value:upper()
		sync_zones(qso)
	elseif field == "name" then
		qso.rx.name = value
	elseif field == "ref" then
		qso.additional.rx_pota = value:upper()

		local res = hlog.sig.lookup_pota(value:upper())
		if res ~= nil then
			qso.rx.qth = res['name']
		end

		-- TODO: fill in more information (state, cqz, ituz, ...)
	else
		error(string.format("Unknown field '%s' encountered", field))
	end

	return nil, {
		["DUP"] = isdup(qso) and "red" or "black",
	}
end

local function startup_fill_template(template)
	local function yesno(msg, default)
		local other = {
			['y'] = 'n',
			['n'] = 'y',
		}
		other = other[default:lower()]

		print(msg .. " [" .. default:upper() .. other .. "]")

		local ack = io.stdin:read()
		if ack == other:lower() or ack == other:upper() then
			return true
		end

		return false
	end

	local function get_state(dxcc, locs)
		local function strip_pfx(loc, pfx)
			return loc:sub(pfx:len() + 1)
		end

		local pfxs = {
			[1] = "CA-", -- Canada
			[6] = "US-", -- Alaska
			[110] = "US-", -- Hawaii
			[291] = "US-", -- continental US
		}

		if pfxs[dxcc] == nil then
			return nil
		end

		if locs:find(",") == nil then
			return strip_pfx(locs, pfxs[dxcc])
		end

		while true do
			print("Which state (e.g., US-MA)?")
			local tmp = io.stdin:read()

			for _, loc in ipairs(hlog.utils.split_on_commas(locs)) do
				assert(loc:find(pfxs[dxcc]) == 1)

				local state = strip_pfx(loc, pfxs[dxcc])

				if tmp == loc or tmp == state then
					return state
				end
			end

			if yesno("Entered state not listed, use anyway?", 'n') then
				return tmp
			end
		end
	end

	local park
	local state

	repeat
		print()
		print("Enter park reference (e.g., K-1234):")

		local ref = io.stdin:read()
		if ref == "" then
			if yesno("Empty park reference, use it anyway?", 'n') then
				return
			end
		else
			park = hlog.sig.lookup_pota(ref)

			if park == nil then
				if yesno("Unknown park reference, use it anyway?", 'n') then
					-- We know nothing but the reference
					template.additional['tx_pota'] = ref
					return
				end
			else
				print("Found:")
				for k,v in pairs(park) do
					print("", k, v)
				end
				print()
				if not park['active'] then
					print("Warning: this reference is marked as inactive!")
					print()
				end

				state = get_state(park['dxcc'], park['locs'])

				if yesno("Use it?", 'y') then
					park = nil
				end
			end
		end
	until park ~= nil

	template.tx.qth = park['name']
	template.tx.dxcc = park['dxcc']
	template.tx.state = state
	template.additional['tx_pota'] = park['ref']

	-- TODO: fill in country, continent, ITU, CQ
end

return {
	labels = {
		{ 0,  0, "St. Call" },
		{ 0, 11, "RSTs" },
		{ 0, 16, "RSTr" },
		{ 0, 21, "St" },
		{ 0, 24, "Name" },
		{ 0, 41, "Refs" },
	},

	fields = {
		{ 1,  0, 10, "rx.station_call", },
		{ 1, 11,  4, "tx.rst", },
		{ 1, 16,  4, "rx.rst", },
		{ 1, 21,  2, "rx.state", },
		{ 1, 24, 16, "name", },
		{ 1, 41, 15, "ref", },
	},

	annunciators = {
		{ 0, 57, "DUP", "black" },
	},

	events = {
		-- called once on startup (after config's fill_template)
		startup = function(template)
			print("POTA contest script")
			startup_fill_template(template)
		end,

		-- called when QSO is being initialized
		--
		-- We rely on the tx_{pota,wwff} additional fields already
		-- being set by the template.
		qso_init = function(qso)
			local ref = qso.additional['tx_pota']

			contest.lcd(ref ~= nil and ref or "?")
		end,

		-- called when QSO is complete & saved, just prior to a reset
		qso_done = qso_done,

		-- called whenever a contest field changes
		field_changed = field_changed,
	},
}
